package es.uoc.mxpdroid.entity;


import es.uoc.mxpdroid.util.GeoUTM;
/**
 * 
 * @author Simone Rondelli - simone.rondelli2@studio.unibo.it
 *
 */
public class XpGeoPoint {
	
	private double latitude;
	private double longitude;
	private double elevation;
	private double resolution;
	
	public XpGeoPoint() {
	}
	
	public XpGeoPoint(double latitude, double longitude) {
		this.latitude = latitude;
		this.longitude = longitude;
	}
	
	public XpGeoPoint(double latitude, double longitude, double altitude) {
		this.latitude = latitude;
		this.longitude = longitude;
		elevation = altitude;
	}

	public XpGeoPoint(double latitude, double longitude, double altitude, double resolution) {
		this.latitude = latitude;
		this.longitude = longitude;
		elevation = altitude;
		this.resolution = resolution;
	}
	
	public void changeEllipsoid(int origEllip, int destEllip){
		double[] geo = {latitude, longitude};
		double[] tmpEFG = (new GeoUTM()).geo2EFG(geo, origEllip);
		geo = (new GeoUTM()).efg2GEO(tmpEFG, destEllip);
		latitude = geo[0];
		longitude = geo[1];
	}
	
	public GridCell getGridCell(RasterLayer layer){
		GeoUTM geo = new GeoUTM(GeoUTM.ETRS89);
		double[] utm = geo.passToUTM(latitude, longitude, 31);
		int[] matrix_coords = layer.get_iijj(utm[0], utm[1]);
		return new GridCell(matrix_coords[0], matrix_coords[1], 100);
	}
	
	public double[] getUTMCoordinates(RasterLayer layer){
		GeoUTM geo = new GeoUTM(GeoUTM.ETRS89);
		return geo.passToUTM(latitude, longitude, 31);
	}
	
	public double getLatitude() {
		return latitude;
	}

	public void setLatitude(double latitude) {
		this.latitude = latitude;
	}

	public double getLongitude() {
		return longitude;
	}

	public void setLongitude(double longitude) {
		this.longitude = longitude;
	}

	public double getElevation() {
		return elevation;
	}

	public void setElevation(double elevation) {
		this.elevation = elevation;
	}

	public double getResolution() {
		return resolution;
	}

	public void setResolution(double resolution) {
		this.resolution = resolution;
	}

	/**
	 * Longitudine in micro-gradi (* 1E6)
	 * 
	 * @return longitude * 1E6
	 */
	public int getMicroLongitude() {
		return (int) (longitude * 1E6);
	}

	/**
	 * Latitudine in micro-gradi (* 1E6)
	 * 
	 * @return latitudine * 1E6
	 */
	public int getMicroLatitude() {
		return (int) (latitude * 1E6);
	}

	@Override
	public boolean equals(Object other) {
		if (other == null) {
			return false;
		}
		if (other instanceof XpGeoPoint) {
			XpGeoPoint oth = (XpGeoPoint) other;
			return oth.getLatitude() == latitude && oth.getLongitude() == longitude && oth.getElevation() == elevation;
		}
		return false;
	}

	@Override
	public String toString() {
		return "GeoPoint [lat=" + latitude + " lon=" + longitude + " ele=" + elevation + "]";
	}
}
